from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()

from sqlalchemy import Column, Integer, String
from sqlalchemy.orm import relationship
from sqlalchemy.schema import ForeignKey


class Timeseries(Base):
    __tablename__ = "timeseries"

    id = Column(Integer, primary_key=True)
    primary_id = Column(String, unique=True)
    ts_instances = relationship("TimeseriesInstance")

    def __repr__(self):
        return "<Timeseries(primary_id='{0}')>".format(self.primary_id)


class Measurand(Base):
    __tablename__ = "measurand"

    id = Column(Integer, primary_key=True)
    short_id = Column(String, unique=True)
    long_id = Column(String, unique=True)
    description = Column(String, unique=True)

    def __repr__(self):
        return "<Measurand(short_id='{0}', long_id='{1}', description={2})>".format(
            self.short_id, self.long_id, self.description
        )


class Source(Base):
    __tablename__ = "source"

    id = Column(Integer, primary_key=True)
    short_id = Column(String, unique=True)
    description = Column(String, unique=True)

    def __repr__(self):
        return "<Source(short_id='{0}', description={1})>".format(
            self.short_id, self.description
        )


class Attribute(Base):
    __tablename__ = "attribute"

    id = Column(Integer, primary_key=True)
    short_id = Column(String, unique=True)
    description = Column(String, unique=True)

    def __repr__(self):
        return "<Attribute(short_id='{0}', description={1})>".format(
            self.short_id, self.description
        )


class AttributeValue(Base):
    __tablename__ = "attribute_value"

    id = Column(Integer, primary_key=True)
    attribute_id = Column(Integer, ForeignKey("attribute.id"))
    attribute_value = Column(String, unique=True)

    def __repr__(self):
        return "<Attribute(attribute_id='{0}', attribute_value={1})>".format(
            self.attribute_id, self.attribute_value
        )


class SchemaVersion(Base):
    __tablename__ = "schema_version"

    id = Column(Integer, primary_key=True)
    version = Column(String)

    def __repr__(self):
        return "<SchemaVersion(version='{0}')>".format(self.version)


class TimeseriesInstance(Base):
    __tablename__ = "timeseries_instance"
    ts_id = Column(Integer, ForeignKey("timeseries.id"), primary_key=True)
    freq = Column(String(10), primary_key=True)
    measurand_id = Column(Integer, ForeignKey("measurand.id"), primary_key=True)
    source_id = Column(Integer, ForeignKey("source.id"), primary_key=True)
    initial_metadata = Column(String(255))
    measurand = relationship("Measurand", backref="measurands")
    timeseries = relationship("Timeseries", backref="timeseries")
    source = relationship("Source", backref="source")
    uuid = Column(String(32))

    def __repr__(self):
        return "<TimeseriesInstance(timeseries='{0}, measurand='{1}', source='{2}')>".format(
            self.timeseries, self.measurand, self.source
        )



class AlreadyExistsError(Exception):
    pass


class DataError(Exception):
    pass


class DuplicateError(Exception):
    pass


class MissingAttributeError(Exception):
    pass


class MissingDataError(Exception):
    pass



import numpy as np
import pandas as pd
import tables
from phildb.constants import MISSING_VALUE, METADATA_MISSING_VALUE


class TabDesc(tables.IsDescription):
    time = tables.Int64Col(dflt=0, pos=0)
    value = tables.Float64Col(dflt=np.nan, pos=1)
    meta = tables.Int32Col(dflt=0, pos=2)
    replacement_time = tables.Int64Col(dflt=0, pos=3)


class LogHandler:
    """
    """

    FILTERS = tables.Filters(complib="zlib", complevel=9)

    def __init__(self, filename, mode):
        self.hdf5 = tables.open_file(filename, mode, filters=self.FILTERS)

    def create_skeleton(self):
        """
            Create the skeleton of the log self.hdf5.
        """
        data_group = self.hdf5.create_group("/", "data", "data group")

        try:
            new_table = self.hdf5.create_table(data_group, "log", TabDesc)
        except tables.exceptions.NodeError as e:
            pass

        self.hdf5.flush()

    def read(self, as_at_datetime):
        field_names = ["time", "value", "meta", "replacement_time"]
        ts_table = self.hdf5.get_node("/data/log")

        records = ts_table.read_where("replacement_time <= {0}".format(as_at_datetime))

        if len(records) == 0:
            return pd.DataFrame(None, columns=field_names)

        df = pd.DataFrame(records, columns=field_names)
        df["date"] = pd.to_datetime(df["time"], unit="s")
        df["replacement_time"] = pd.to_datetime(df["replacement_time"], unit="s")
        df = df.set_index("date")
        df.drop("time", axis=1, inplace=True)

        meta_ids = df.meta.copy()
        replacement_times = df.replacement_time.copy()
        df.loc[df.meta == METADATA_MISSING_VALUE] = np.nan
        df.meta = meta_ids
        df.replacement_time = replacement_times

        idx = ~df.index.duplicated(keep="last")

        df = df.loc[idx]

        return df

    def write(self, log_entries, operation_datetime):

        ts_table = self.hdf5.get_node("/data/log")

        index_row = ts_table.row
        for dt, val, meta in iter(log_entries["C"]):
            if val is np.nan:
                val = MISSING_VALUE
                meta = METADATA_MISSING_VALUE

            index_row["time"] = dt
            index_row["value"] = val
            index_row["meta"] = meta
            index_row["replacement_time"] = operation_datetime
            index_row.append()

        self.hdf5.flush()

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.__del__()

    def __del__(self):
        if self.hdf5 is not None:
            self.hdf5.close()
            self.hdf5 = None

    def __str__(self):
        return str(self.hdf5)



import argparse
import os

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker

Session = sessionmaker()

from phildb import constants
from phildb import dbstructures
from phildb.exceptions import AlreadyExistsError


def create(tsdb_path):

    if not os.path.exists(tsdb_path):
        os.makedirs(tsdb_path)
        os.makedirs(os.path.join(tsdb_path, "data"))
    elif not os.listdir(tsdb_path):
        os.makedirs(os.path.join(tsdb_path, "data"))
    else:
        raise AlreadyExistsError(
            "PhilDB database already exists at: {0}".format(tsdb_path)
        )

    engine = create_engine(
        "sqlite:///{0}{1}{2}".format(tsdb_path, os.path.sep, constants.METADATA_DB)
    )
    dbstructures.Base.metadata.create_all(engine)

    Session.configure(bind=engine)
    session = Session()

    version = dbstructures.SchemaVersion(version=constants.DB_VERSION)
    session.add(version)
    session.commit()


def main():
    parser = argparse.ArgumentParser(description="Create PhilDB database.")
    parser.add_argument("dbname", help="PhilDB database to create")

    args = parser.parse_args()

    create(args.dbname)


if __name__ == "__main__":
    main()



import argparse

# Disable unused import warnings, because we want these imported
# not for immediate use but for interactive use once the console
# is loaded.
from datetime import datetime, date  # pylint:disable=unused-import
import numpy as np  # pylint:disable=unused-import
import pandas as pd  # pylint:disable=unused-import

try:
    from matplotlib import pyplot as plt  # pylint:disable=unused-import
except ImportError:
    # Having matplotlib isn't a hard dependency, so don't worry if it wasn't imported.
    pass

from IPython.terminal.embed import InteractiveShellEmbed

from phildb import __version__
from phildb.database import PhilDB


def main(deprecated=False):
    ipshell = InteractiveShellEmbed()

    parser = argparse.ArgumentParser(description="Open PhilDB database.")
    parser.add_argument("dbname", help="PhilDB database to open", nargs="?")
    parser.add_argument(
        "--version", action="store_true", help="Print version and exit."
    )

    args = parser.parse_args()

    if args.version:
        print("PhilDB version: {0}".format(__version__))
        exit()

    if args.dbname is None:
        print(parser.print_help())
        exit()

    if deprecated:
        deprecation_warning = "DEPRECATION WARNING: The `phil` command has been deprecated, use `phildb` instead.\n\n"
    else:
        deprecation_warning = ""

    db = PhilDB(args.dbname)
    ipshell(
        "{1}"
        "Running timeseries database: {0}\n"
        "Access the 'db' object to operate on the database.\n"
        "\n"
        "Run db.help() for a list of available commands.".format(
            db, deprecation_warning
        )
    )


def deprecated_main():
    main(True)


if __name__ == "__main__":
    main()



from datetime import datetime
import hashlib
import os
import types
import uuid

import pandas as pd

from sqlalchemy import create_engine
from sqlalchemy.exc import IntegrityError
from sqlalchemy.orm import sessionmaker, joinedload
from sqlalchemy.orm.exc import NoResultFound

import logging

logger = logging.getLogger("PhilDB_database")

from phildb import constants
from phildb import reader
from phildb import writer
from phildb.dbstructures import SchemaVersion, Timeseries, Measurand, TimeseriesInstance
from phildb.dbstructures import Source
from phildb.dbstructures import Attribute, AttributeValue
from phildb.exceptions import DuplicateError, MissingAttributeError, MissingDataError


class PhilDB(object):
    def __init__(self, tsdb_path):
        self.tsdb_path = tsdb_path

        logger.debug(self.__meta_data_db())

        if not os.path.exists(self.tsdb_path):
            raise IOError("PhilDB database doesn't exist ({0})".format(self.tsdb_path))

        if not os.path.exists(self.__meta_data_db()):
            raise IOError(
                "PhilDB database doesn't contain meta-database ({0})".format(
                    self.__meta_data_db()
                )
            )

        self.__engine = create_engine("sqlite:///{0}".format(self.__meta_data_db()))
        self.Session = sessionmaker()
        self.Session.configure(bind=self.__engine)

        assert self.version() == constants.DB_VERSION

    def __meta_data_db(self):
        return os.path.join(self.tsdb_path, constants.METADATA_DB)

    def __data_dir(self):
        return os.path.join(self.tsdb_path, "data")

    def help(self):
        """
            List methods of the PhilDB class with the first line of their docstring.
        """
        for method in sorted(dir(self)):
            if method.startswith("_"):
                continue
            if not isinstance(getattr(self, method), types.MethodType):
                continue

            docstring = getattr(self, method).__doc__
            if docstring is None:
                short_string = ""
            else:
                docstring = docstring.split("\n")
                short_string = docstring[0].strip()
                if short_string == "":
                    short_string = docstring[1].strip()
            print("{0}: {1}".format(method, short_string))

    def version(self):
        """
            Returns the version number of the database schema.

            :returns: string -- Schema version.
        """
        session = self.Session()
        query = session.query(SchemaVersion.version)

        version = query.scalar()

        return version

    def add_timeseries(self, identifier):
        """
            Create a timeseries entry to be identified by the supplied ID.

            :param identifier: Identifier of the timeseries.
            :type identifier: string
        """
        the_id = identifier.strip()
        session = self.Session()
        ts = Timeseries(primary_id=the_id)
        session.add(ts)
        try:
            session.commit()
        except IntegrityError:
            raise DuplicateError("Already exists: '{0}'".format(the_id))

    def add_measurand(self, measurand_short_id, measurand_long_id, description):
        """
            Create a measurand entry.

            Measurand being a measurable timeseries type. e.g. Streamflow, Temperature, Rainfall, etc.

            :param measurand_short_id: Short identifier of the measurand.
            :type measurand_short_id: string
            :param measurand_long_id: Long identifier of the measurand.
            :type measurand_long_id: string
            :param description: Description of the measurand.
            :type description: string
        """
        short_id = measurand_short_id.strip()
        long_id = measurand_long_id.strip()
        session = self.Session()
        measurand = Measurand(
            short_id=short_id, long_id=long_id, description=description
        )
        session.add(measurand)
        try:
            session.commit()
        except IntegrityError:
            raise DuplicateError("Already exists: '{0}'".format(measurand_short_id))

    def add_source(self, source, description):
        """
            Define a source.

            Source being the origin of the data. For example the source used in
            the examples/hrs example is BOM_HRS. Indicated the origin of the
            data was the Bureau of Metorology Hydrologic Reference Stations
            project.

            :param source: Identifier of the source.
            :type source: string
            :param description: Description of the source.
            :type description: string
        """
        short_id = source.strip()
        session = self.Session()
        source = Source(short_id=short_id, description=description)
        session.add(source)
        try:
            session.commit()
        except IntegrityError:
            raise DuplicateError("Already exists: '{0}'".format(source))

    def add_attribute(self, attribute_id, description):
        """
            Define an attribute.

            :param attribute_id: Identifier of the attribute.
            :type attribute_id: string
            :param description: Description of the attribute.
            :type description: string
        """
        short_id = attribute_id.strip().upper()
        session = self.Session()
        attribute = Attribute(short_id=short_id, description=description)
        session.add(attribute)
        session.commit()

    def add_attribute_value(self, attribute_id, value):
        """
            Store an attribute value.

            :param attribute_id: Identifier of the attribute.
            :type attribute_id: string
            :param value: The attribute value to store.
            :type value: string
        """
        short_id = attribute_id.strip().upper()
        session = self.Session()

        query = session.query(Attribute).filter(Attribute.short_id == short_id)
        try:
            attribute = query.one()
        except NoResultFound as e:
            raise MissingAttributeError(
                "Could not find {0} ({1}) in the database.".format(attribute_id, value)
            )

        attribute = AttributeValue(attribute_id=attribute.id, attribute_value=value)
        session.add(attribute)
        session.commit()

    def __parse_attribute_kwargs(self, **kwargs):
        """
            Convert kwargs of attribute short IDs into database objects

            :param **kwargs: Any additional attributes to attach to the timeseries instance.
            :type \*\*kwargs: kwargs

            :return: Dictionary of attributes from the database.
        """
        session = kwargs.pop("session", None)
        attributes = {}
        for attribute, value in kwargs.items():
            if value is None:
                continue
            attributes[attribute] = self.__get_attribute(attribute, value, session)

        return attributes

    def add_timeseries_instance(self, identifier, freq, initial_metadata, **kwargs):
        """
            Define an instance of a timeseries.

            A timeseries instance is a combination of a timeseries, frequency and attributes.

            :param identifier: Identifier of the timeseries.
            :type identifier: string
            :param freq: Data frequency (e.g. 'D' for day, as supported by pandas.)
            :type freq: string
            :param initial_metadata: Store some metadata about this series.
                Potentially freeform header from a source file about to be loaded.
            :type initial_metadata: string
            :param \*\*kwargs: Any additional attributes to attach to the timeseries instance.
            :type \*\*kwargs: kwargs
        """
        session = self.Session()

        timeseries = self.__get_record_by_id(identifier, session)

        attributes = self.__parse_attribute_kwargs(session=session, **kwargs)

        query = session.query(TimeseriesInstance).filter_by(
            timeseries=timeseries, freq=freq, **attributes
        )
        try:
            record = query.one()
            session.rollback()
            raise DuplicateError(
                "TimeseriesInstance for ({:}) already exists.".format(
                    identifier, **kwargs
                )
            )
        except NoResultFound as e:
            # No result is good, we can now create a ts instance.
            pass

        with session.no_autoflush:
            tsi = TimeseriesInstance(initial_metadata=initial_metadata)
            tsi.measurand = attributes["measurand"]
            tsi.source = attributes["source"]
            tsi.freq = freq
            tsi.uuid = uuid.uuid4().hex
            timeseries.ts_instances.append(tsi)

            session.add(tsi)
            try:
                session.commit()
            except IntegrityError:
                raise DuplicateError(
                    "Timeseries instance already exists: '{0}', '{1}'".format(
                        identifier, freq
                    )
                )

    def __get_record_by_id(self, identifier, session=None):
        """
            Get a database record for the given timeseries ID.

            If a timeseries record can not be found a MissingDataError is raised.

            :param identifier: Identifier of the timeseries.
            :type identifier: string
            :param session: Database session to use. (Optional)
            :type session: sqlalchemy.orm.sessionmaker.Session
            :returns: Single session.query result.
            :raises: MissingDataError
        """
        if session is None:
            session = self.Session()

        query = session.query(Timeseries).filter(Timeseries.primary_id == identifier)
        try:
            record = query.one()
        except NoResultFound as e:
            raise MissingDataError(
                "Could not find metadata record for: {0}".format(identifier)
            )

        return record

    def __get_attribute(self, attribute, value, session=None):
        """
            Get a database record for the given attribute.

            If a attribute record can not be found a MissingAttributeError is raised.

            :param attribute: Identifier of the attribute.
            :type attribute: string
            :param session: Database session to use. (Optional)
            :type session: sqlalchemy.orm.sessionmaker.Session
            :returns: Single session.query result.
            :raises: MissingAttributeError
        """
        if session is None:
            session = self.Session()

        if attribute == "measurand":
            query = session.query(Measurand).filter(Measurand.short_id == value)
        elif attribute == "source":
            query = session.query(Source).filter(Source.short_id == value)
        elif attribute == "timeseries":
            query = session.query(Timeseries).filter(Timeseries.primary_id == value)
        elif attribute == "provider":
            short_id = attribute.strip().upper()
            query = session.query(Attribute).filter(Attribute.short_id == short_id)
            try:
                record = query.one()
            except NoResultFound as e:
                raise MissingAttributeError(
                    "Could not find {0} ({1}) in the database.".format(attribute, value)
                )
            query = session.query(AttributeValue).filter(
                AttributeValue.attribute_id == record.id,
                AttributeValue.attribute_value == value,
            )
        else:
            raise MissingAttributeError("Attribute {0} unknown".format(attribute))

        try:
            record = query.one()
        except NoResultFound as e:
            raise MissingAttributeError(
                "Could not find {0} ({1}) in the database.".format(attribute, value)
            )

        return record

    def get_file_path(self, identifier, freq, ftype="tsdb", **kwargs):
        """
            Get a path to a file for a given timeseries instance.

            :param identifier: Identifier of the timeseries.
            :type identifier: string
            :param ftype: File extension to use (i.e. the type of file).
                (Default='tsdb')
            :type ftype: string
            :returns: string -- Path to file for a timeseries instance identified
                by the given arguments.
        """
        record = self.__get_ts_instance(identifier, freq, **kwargs)

        return os.path.join(self.__data_dir(), record.uuid + "." + ftype)

    def write(self, identifier, freq, ts, **kwargs):
        """
            Write/update timeseries data for existing timeseries.

            :param identifier: Identifier of the timeseries.
            :type identifier: string
            :param freq: Data frequency (e.g. 'D' for day, as supported by pandas.)
            :type freq: string
            :param ts: Timeseries data to write into the database.
            :type ts: pd.Series
        """
        modified = writer.write(
            self.get_file_path(identifier, freq, **kwargs), ts, freq
        )

        log_file = self.get_file_path(identifier, freq, ftype="hdf5", **kwargs)

        writer.write_log(log_file, modified, datetime.utcnow())

    def read(self, identifier, freq, **kwargs):
        """
            Read the entire timeseries record for the requested timeseries instance.

            :param identifier: Identifier of the timeseries.
            :type identifier: string
            :param freq: Timeseries data frequency.
            :type freq: string
            :param kwargs: Attributes to match against timeseries instances (e.g. source, measurand).
            :type kwargs: kwargs

            :returns: pandas.DataFrame -- Timeseries data.
        """
        return reader.read(self.get_file_path(identifier, freq, **kwargs))

    def read_log(self, identifier, freq, as_at_datetime, **kwargs):
        """
            Read timeseries record for the requested timeseries instance as it was at specified datetime in the log.

            :param identifier: Identifier of the timeseries.
            :type identifier: string
            :param freq: Timeseries data frequency.
            :type freq: string
            :param as_at_datetime: Filter to a timeseries, as available at this specified datetime, from the log.
            :type as_at_datetime: datetime
            :param kwargs: Attributes to match against timeseries instances (e.g. source, measurand).
            :type kwargs: kwargs

            :returns: pandas.DataFrame -- Timeseries data.
        """
        return reader.read_log(
            self.get_file_path(identifier, freq, ftype="hdf5", **kwargs), as_at_datetime
        )

    def read_all(self, freq, excludes=None, **kwargs):
        """
            Read the entire timeseries record for all matching timeseries instances.
            Optionally exclude timeseries from the final DataFrame by specifying IDs in the exclude argument.

            :param identifier: Identifier of the timeseries.
            :type identifier: string
            :param freq: Timeseries data frequency.
            :type freq: string
            :param excludes: IDs of timeseries to exclude from final DataFrame.
            :type excludes: array[string]
            :param kwargs: Attributes to match against timeseries instances (e.g. source, measurand).
            :type kwargs: kwargs

            :returns: pandas.DataFrame -- Timeseries data.
        """
        data = {}
        if excludes is None:
            identifiers = self.ts_list(**kwargs)
        else:
            identifiers = set(self.ts_list(**kwargs)).difference(excludes)

        return self.read_dataframe(identifiers, freq, **kwargs)

    def read_dataframe(self, identifiers, freq, **kwargs):
        """
            Read the entire timeseries record for the requested timeseries instances.

            :param identifiers: Identifiers of the timeseries to read into a DataFrame.
            :type identifiers: array[string]
            :param freq: Timeseries data frequency.
            :type freq: string
            :param kwargs: Attributes to match against timeseries instances (e.g. source, measurand).
            :type kwargs: kwargs

            :returns: pandas.DataFrame -- Timeseries data.
        """
        data = {}
        for ts_id in identifiers:
            data[ts_id] = reader.read(self.get_file_path(ts_id, freq, **kwargs))
        return pd.DataFrame(data)

    def ts_list(self, **kwargs):
        """
            Returns list of primary ID for all timeseries records.

            :param kwargs: Restrict to records associated with this the kwargs
                attributes supplied. (Optional).
            :type kwargs: kwargs
            :returns: list(string) -- Sorted list of timeseries identifiers.
        """
        session = self.Session()
        query_args = self.__parse_attribute_kwargs(**kwargs)

        records = (
            session.query(TimeseriesInstance)
            .options(joinedload(TimeseriesInstance.timeseries))
            .filter_by(**query_args)
        )
        return sorted(list(set([record.timeseries.primary_id for record in records])))

    def list_ids(self):
        """
            Returns list of timeseries IDs for all timeseries records.

            :returns: list(string) -- Sorted list of timeseries identifiers.
        """
        session = self.Session()

        records = session.query(Timeseries)
        return sorted(list(set([record.primary_id for record in records])))

    def list_timeseries_instances(self, **kwargs):
        """
            Returns list of timeseries instances for all instance records.

            Can filter by using keyword arguments.

            :returns: list(string) -- Sorted list of timeseries instances.
        """
        session = self.Session()

        initial_args = {}
        for attr in ["freq"]:
            attr_val = kwargs.pop(attr, None)

            if attr_val:
                initial_args[attr] = attr_val

        query_args = self.__parse_attribute_kwargs(**kwargs)
        query_args.update(initial_args)

        records = (
            session.query(TimeseriesInstance)
            .options(joinedload(TimeseriesInstance.timeseries))
            .filter_by(**query_args)
        )
        instance_list = []
        for record in records:
            instance = {
                "ts_id": record.timeseries.primary_id,
                "freq": record.freq,
                "measurand": record.measurand.short_id,
                "source": record.source.short_id,
            }
            instance_list.append(instance)

        return pd.DataFrame(instance_list)

    def list_measurands(self):
        """
            Returns list of measurand short IDs for all measurand records.

            :returns: list(string) -- Sorted list of timeseries identifiers.
        """
        session = self.Session()

        records = session.query(Measurand)
        return sorted(list(set([record.short_id for record in records])))

    def list_sources(self):
        """
            Returns list of source IDs for all sources.

            :returns: list(string) -- Sorted list of source identifiers.
        """
        session = self.Session()

        records = session.query(Source)
        return sorted(list(set([record.short_id for record in records])))

    def read_metadata(self, ts_id, freq, **kwargs):
        """
            Returns the metadata that was associated with an initial TimeseriesInstance.

            :param identifier: Identifier of the timeseries.
            :type identifier: string
            :returns: string -- The initial metadata that was recorded on
                instance creation.
        """
        return self.__get_ts_instance(ts_id, freq, **kwargs).initial_metadata

    def __get_ts_instance(self, ts_id, freq, **kwargs):
        """
            Get a database record for the requested timeseries instance.

            If a timeseries instance record can not be found a MissingDataError is
            raised.

            :param ts_id: Identifier of the timeseries.
            :type ts_id: string
            :returns: dbstructures.TimeseriesInstance -- Single session.query result.
            :raises: MissingDataError
        """
        timeseries = self.__get_record_by_id(ts_id)

        query_args = self.__parse_attribute_kwargs(**kwargs)

        session = self.Session()
        query = session.query(TimeseriesInstance).filter_by(
            timeseries=timeseries, freq=freq, **query_args
        )

        try:
            record = query.one()
        except NoResultFound as e:
            raise MissingDataError(
                "Could not find TimeseriesInstance for ({:}).".format(
                    ts_id, freq, **kwargs
                )
            )

        return record

    def __str__(self):
        return self.tsdb_path



import calendar
from datetime import datetime as dt, timedelta
from dateutil.relativedelta import relativedelta
import numpy as np
import os
import pandas as pd
from struct import pack, unpack, calcsize

import logging

logger = logging.getLogger(__name__)

from phildb.constants import DEFAULT_META_ID, MISSING_VALUE, METADATA_MISSING_VALUE
from phildb.log_handler import LogHandler
from phildb.exceptions import DataError
from phildb.reader import __read, read

field_names = ["date", "value", "metaID"]
entry_format = "<qdi"  # long, double, int; See field names above.
entry_size = calcsize(entry_format)


def __pack(record_date, value, default_flag=DEFAULT_META_ID):

    if np.isnan(value):
        data = pack(entry_format, record_date, MISSING_VALUE, METADATA_MISSING_VALUE)
    else:
        data = pack(entry_format, record_date, value, default_flag)

    return data


def __convert_and_validate(ts, freq):
    """
        Enforces frequency.

        :param ts: Pandas Series object
        :type ts: pandas.Series
    """

    ts.sort_index(inplace=True)

    if freq == "IRR":
        series = ts
    else:
        series = ts.asfreq(freq)

    return series


def __calculate_offset(freqstr, start_date, first_record_date):
    if freqstr[-1] == "T":
        if len(freqstr) == 1:
            freq_mult = 1
        else:
            freq_mult = int(freqstr[:-1])
        freqstr = "T"
    else:
        freq_mult = 1

    if freqstr[-1] == "S" and len(freqstr) > 1:
        freqstr = freqstr[:-1]

    return (
        start_date.to_period(freqstr)
        - pd.to_datetime(first_record_date).to_period(freqstr)
    ).n


def __write_missing(writer, freq, first_date, last_date, log_entries):
    log_entries = log_entries.copy()

    missing_dates = pd.date_range(first_date, last_date, freq=freq)
    for the_date in missing_dates:
        datestamp = calendar.timegm(the_date.utctimetuple())
        log_entries["C"].append((datestamp, MISSING_VALUE, METADATA_MISSING_VALUE))

        data = pack(entry_format, datestamp, MISSING_VALUE, METADATA_MISSING_VALUE)

        writer.write(data)

    return log_entries


def __update_existing_data(tsdb_file, series, log_entries):

    log_entries = log_entries.copy()
    with open(tsdb_file, "rb") as reader:
        first_record = unpack(entry_format, reader.read(entry_size))

    first_record_date = dt(1970, 1, 1) + timedelta(seconds=first_record[0])

    offset = __calculate_offset(
        series.index.freqstr, series.index[0], first_record_date
    )

    with open(tsdb_file, "r+b") as writer:
        existing_records = []

        # Read existing overlapping data for comparisons
        writer.seek(entry_size * offset, os.SEEK_SET)

        for record in iter(lambda: writer.read(entry_size), ""):
            if not record:
                break
            existing_records.append(unpack(entry_format, record))

        records_length = len(existing_records)

        # Start a count for records from the starting write position
        rec_count = 0
        writer.seek(entry_size * offset, os.SEEK_SET)
        for date, value in zip(series.index, series.values):
            datestamp = calendar.timegm(date.utctimetuple())
            overlapping = rec_count <= records_length - 1

            if overlapping and (
                existing_records[rec_count][1] == value
                or existing_records[rec_count][2] == MISSING_VALUE
            ):
                # Skip writing the entry if it hasn't changed.
                writer.seek(
                    entry_size * (rec_count + 1) + (entry_size * offset), os.SEEK_SET
                )
            elif overlapping and existing_records[rec_count][1] != value:
                log_entries["U"].append(existing_records[rec_count])

                log_entries["C"].append((datestamp, value, DEFAULT_META_ID))
                data = __pack(datestamp, value)
                writer.write(data)
            else:
                data = __pack(datestamp, value)
                log_entries["C"].append((datestamp, value, DEFAULT_META_ID))
                writer.write(data)
            rec_count += 1

    return log_entries


def write(tsdb_file, ts, freq):
    """
        Smart write.

        Will only update existing values where they have changed.
        Changed existing values are returned in a list.

        :param tsdb_file: File to write timeseries data into.
        :type tsdb_file: string
        :param ts: Timeseries data to write.
        :type ts: pd.Series
        :param freq: Frequency of the data. (e.g. 'D' for daily, '1Min' for minutely).
            Accepts any string that pandas.TimeSeries.asfreq does or 'IRR' for irregular data.
        :type freq: string
    """

    series = __convert_and_validate(ts, freq)

    log_entries = {"C": [], "U": []}

    if len(series) == 0:
        return log_entries

    # If the file didn't exist it is a straight foward write and we can
    # just return at the end of this if block.
    if not os.path.isfile(tsdb_file):
        with open(tsdb_file, "wb") as writer:
            for date, value in zip(series.index, series.values):
                datestamp = calendar.timegm(date.utctimetuple())
                log_entries["C"].append((datestamp, value, DEFAULT_META_ID))
                data = __pack(datestamp, value)
                writer.write(data)

        return log_entries

    # If we reached here it wasn't a straight write to a new file.
    if freq == "IRR":
        return write_irregular_data(tsdb_file, series)
    else:
        return write_regular_data(tsdb_file, series)


def write_regular_data(tsdb_file, series):
    """
        Smart write. Expects continuous time series.

        Will only update existing values where they have changed.
        Changed existing values are returned in a list.

        :param tsdb_file: File to write timeseries data into.
        :type tsdb_file: string
        :param series: Pandas Series of regular data to write.
        :type series: pandas.Series
    """
    start_date = series.index[0]
    end_date = series.index[-1]

    log_entries = {"C": [], "U": []}
    with open(tsdb_file, "rb") as reader:
        first_record = unpack(entry_format, reader.read(entry_size))
        reader.seek(entry_size * -1, os.SEEK_END)
        last_record = unpack(entry_format, reader.read(entry_size))

    first_record_date = pd.Timestamp(
        dt(1970, 1, 1) + timedelta(seconds=first_record[0]), freq=series.index.freq
    )
    last_record_date = pd.Timestamp(
        dt(1970, 1, 1) + timedelta(seconds=last_record[0]), freq=series.index.freq
    )

    offset = __calculate_offset(series.index.freqstr, start_date, first_record_date)

    # We are prepending to existing data
    if start_date < first_record_date:
        # Keep original data before writing (since a new file needs to be
        # written for a prepend operation)
        os.rename(tsdb_file, tsdb_file + ".tmp")

        # Write all the data up to the original first_record_date
        with open(tsdb_file, "wb") as writer:
            for date, value in zip(
                series.loc[: first_record_date - series.index.freq].index,
                series.loc[: first_record_date - series.index.freq].values,
            ):
                datestamp = calendar.timegm(date.utctimetuple())
                log_entries["C"].append((datestamp, value, DEFAULT_META_ID))
                data = __pack(datestamp, value)
                writer.write(data)

        # Fill any missing values between the end of the new series and the start of the old
        with open(tsdb_file, "ab") as writer:
            log_entries = __write_missing(
                writer,
                series.index.freq,
                end_date + series.index.freq,
                pd.Timestamp(first_record_date, freq=series.index.freq)
                - series.index.freq,
                log_entries,
            )

        # Copy over existing data
        with open(tsdb_file + ".tmp", "rb") as original_data, open(
            tsdb_file, "ab"
        ) as writer:
            for chunk in iter(lambda: original_data.read(entry_size), b""):
                writer.write(chunk)

        # Remove temporary file now that its contents has been copied
        os.remove(tsdb_file + ".tmp")

        # Update existing data
        if len(series.loc[first_record_date:]) > 0:
            log_entries = __update_existing_data(
                tsdb_file, series.loc[first_record_date:], log_entries
            )

    # We are updating existing data
    elif start_date <= last_record_date:
        log_entries = __update_existing_data(tsdb_file, series, log_entries)

    # We are appending data
    elif start_date > last_record_date:
        with open(tsdb_file, "a+b") as writer:
            last_record_date = (
                pd.Timestamp(last_record_date, unit=series.index.freq.name)
                + series.index.freq
            )

            log_entries = __write_missing(
                writer,
                series.index.freq,
                last_record_date,
                start_date - series.index.freq,
                log_entries,
            )

            for date, value in zip(series.index, series.values):
                datestamp = calendar.timegm(date.utctimetuple())
                log_entries["C"].append((datestamp, value, DEFAULT_META_ID))
                data = __pack(datestamp, value)
                writer.write(data)

    else:  # Not yet supported
        raise NotImplementedError

    return log_entries


def write_irregular_data(tsdb_file, series):
    """
        Smart write of irregular data.

        Will only update existing values where they have changed.
        Changed existing values are returned in a list.

        :param tsdb_file: File to write timeseries data into.
        :type tsdb_file: string
        :param series: Pandas Series of irregular data to write.
        :type series: pandas.Series
        :type freq: string
    """
    existing = __read(tsdb_file)

    if series.dtype == np.float32:
        series = series.astype(np.float64)

    overlap_idx = existing.index.intersection(series.index)
    modified = series.loc[overlap_idx] != existing.value.loc[overlap_idx]
    records_to_modify = existing.loc[overlap_idx].loc[modified.values]
    new_records = series.index.difference(existing.index)

    log_entries = {"C": [], "U": []}

    if len(records_to_modify) == 0 and len(new_records) == 0:
        return log_entries

    for date, orig_value, meta_id, new_value in zip(
        records_to_modify.index,
        records_to_modify.value,
        records_to_modify.metaID,
        series.loc[records_to_modify.index],
    ):
        datestamp = int(date.value / 1000000000)
        log_entries["C"].append((datestamp, new_value, meta_id))
        log_entries["U"].append((datestamp, orig_value, meta_id))

    append_only = len(overlap_idx) == 0 and existing.index[-1] < series.index[0]
    if append_only:
        merged = series
        fmode = "ab"
    else:
        # combine_first does not preserve null values in the original series.
        # So do an initial merge.
        merged = series.combine_first(existing.value)
        fmode = "wb"

    # Then replace the null values from the update series.
    null_vals = series.isnull()
    null_idx = null_vals.loc[null_vals == True].index
    merged.loc[null_idx] = np.nan

    merged = pd.DataFrame({"value": merged})

    merged["datestamp"] = pd.Series(
        merged.index.map(lambda dateval: dateval.value // 1000000000),
        index=merged.index,
    )

    new_logs = [
        (row.datestamp, row.value, DEFAULT_META_ID)
        for idx, row in merged.loc[new_records].iterrows()
    ]
    log_entries["C"] += new_logs

    # Only attempt to do backups of file if we're doing a destructive
    # write. This means if we are opening the file in "w" mode. If in
    # append_only then we are using "a" mode which is non-destructive
    # and we really actually want to keep the existing file's data.
    if not append_only:
        os.rename(tsdb_file, tsdb_file + "backup")

    try:
        with open(tsdb_file, fmode) as writer:

            def write_record(row):
                writer.write(__pack(int(row[0]), row[1]))
                return row

            np.apply_along_axis(write_record, 1, merged[["datestamp", "value"]].values)
    except Exception:
        # On any failure writing restore the original file.
        if not append_only:
            os.rename(tsdb_file + "backup", tsdb_file)
        logger.exception(
            "Error writing irregular data to %s. No data change made.", tsdb_file
        )
        raise
    else:
        # On successfull write remove the original file.
        if not append_only:
            os.unlink(tsdb_file + "backup")

    return log_entries


def write_log(log_file, modified, replacement_datetime):

    if not os.path.exists(log_file):
        with LogHandler(log_file, "w") as writer:
            writer.create_skeleton()

    with LogHandler(log_file, "a") as writer:
        writer.write(modified, calendar.timegm(replacement_datetime.utctimetuple()))



import os
from io import open

import versioneer

from setuptools import setup

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "README.rst"), encoding="utf-8") as f:
    long_description = f.read()

requirements = [
    "ipython>=2.0.0",
    "numpy>=1.8.0",
    "pandas>=0.24.2",
    "SQLAlchemy>=0.9.2",
    "tables>=3.1.0",
]

setup(
    name="PhilDB",
    version=versioneer.get_version(),
    cmdclass=versioneer.get_cmdclass(),
    description="Timeseries database",
    long_description=long_description,
    author="Andrew MacDonald",
    author_email="andrew@maccas.net",
    license="BSD",
    url="https://github.com/amacd31/phildb",
    install_requires=requirements,
    packages=["phildb"],
    test_suite="nose.collector",
    tests_require=["nose", "mock"],
    entry_points={
        "console_scripts": [
            "phil-create = phildb.create:main",
            "phildb = phildb.console:main",
            "phil = phildb.console:deprecated_main",
        ]
    },
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Intended Audience :: Science/Research",
        "Topic :: Software Development :: Build Tools",
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: BSD License",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
    ],
)



import calendar
from struct import unpack, calcsize
import numpy as np
import pandas as pd
import os

from phildb.constants import METADATA_MISSING_VALUE
from phildb.log_handler import LogHandler


def __read(filename):
    field_names = ["date", "value", "metaID"]
    entry_format = "<qdi"  # long, double, int; See field names above.
    entry_size = calcsize(entry_format)

    if not os.path.exists(filename):
        return pd.DataFrame(None, columns=["date", "value", "metaID"])

    records = np.fromfile(
        filename, dtype=np.dtype({"names": field_names, "formats": entry_format[1:]})
    )

    if len(records) == 0:
        return pd.DataFrame(None, columns=["date", "value", "metaID"])

    df = pd.DataFrame(records, columns=field_names)
    df["date"] = pd.to_datetime(df["date"], unit="s")
    df = df.set_index("date")

    meta_ids = df.metaID
    df.loc[df.metaID == METADATA_MISSING_VALUE] = np.nan
    df.metaID = meta_ids

    return df


def read(filename):
    return __read(filename).value


def read_log(log_file, as_at_datetime):

    with LogHandler(log_file, "r") as reader:
        df = reader.read(calendar.timegm(as_at_datetime.utctimetuple()))

    return df.value
